/*
 * Copyright 2006-2008 The FLWOR Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <vector>

#include "compiler/expression/fo_expr.h"
#include "compiler/expression/expr_visitor.h"

#include "context/static_context.h"

#include "functions/library.h"
#include "functions/function.h"

#include "zorbaerrors/Assert.h"

namespace zorba
{

SERIALIZABLE_CLASS_VERSIONS(fo_expr)
END_SERIALIZABLE_CLASS_VERSIONS(fo_expr)

DEF_EXPR_ACCEPT (fo_expr)



/*******************************************************************************
  first-order expressions. Represents function invocations as well as:

  OrExpr, AndExpr, ComparisonExpr, RangeExpr, AdditiveExpr, MultiplicativeExpr,
  UnionExpr, and IntersectExceptExpr.
********************************************************************************/

fo_expr* fo_expr::create_seq(static_context* sctx, const QueryLoc& loc)
{
  function* f = BuiltinFunctionLibrary::getFunction(FunctionConsts::OP_CONCATENATE_N);

  std::auto_ptr<fo_expr> fo(new fo_expr(sctx, loc, f));

  return fo.release();
}


fo_expr::fo_expr(
    static_context* sctx,
    const QueryLoc& loc,
    const function* f)
  :
  expr(sctx, loc, fo_expr_kind),
  theFunction(const_cast<function*>(f))
{
  assert(f != NULL);
  compute_scripting_kind();
  setDirectAnnotations();
}


fo_expr::fo_expr(
    static_context* sctx,
    const QueryLoc& loc,
    const function* f,
    const expr* arg)
  :
  expr(sctx, loc, fo_expr_kind),
  theFunction(const_cast<function*>(f))
{
  assert(f != NULL);
  theArgs.resize(1);
  theArgs[0] = arg;

  compute_scripting_kind();
  setDirectAnnotations();
}


fo_expr::fo_expr(
    static_context* sctx,
    const QueryLoc& loc,
    const function* f,
    const expr* arg1,
    const expr* arg2)
  :
  expr(sctx, loc, fo_expr_kind),
  theFunction(const_cast<function*>(f))
{
  assert(f != NULL);
  theArgs.resize(2);
  theArgs[0] = arg1;
  theArgs[1] = arg2;

  compute_scripting_kind();
  setDirectAnnotations();
}


fo_expr::fo_expr(
    static_context* sctx,
    const QueryLoc& loc,
    const function* f,
    const std::vector<expr_t>& args)
  :
  expr(sctx, loc, fo_expr_kind),
  theArgs(args),
  theFunction(const_cast<function*>(f))
{
  assert(f != NULL);
  compute_scripting_kind();
  setDirectAnnotations();
}


void fo_expr::serialize(::zorba::serialization::Archiver &ar)
{
  serialize_baseclass(ar, (expr*)this);
  ar & theArgs;
  SERIALIZE_FUNCTION((function*&)theFunction);
}


const signature& fo_expr::get_signature() const
{
  return theFunction->getSignature();
}


const store::Item* fo_expr::get_fname() const
{
  return theFunction->getName();
}


void fo_expr::compute_scripting_kind()
{
  const function* func = get_func();
  ulong numArgs = num_args();

  if (func->getKind() == FunctionConsts::OP_CONCATENATE_N)
  {
    expr_script_kind_t kind = VACUOUS_EXPR;

    for (ulong i = 0; i < numArgs; ++i)
    {
      if (theArgs[i] == NULL)
        continue;

      expr_script_kind_t argKind = theArgs[i]->get_scripting_kind();

      kind = scripting_kind_anding(kind, argKind, theArgs[i]->get_loc());
    }

    theScriptingKind = kind;
  }
  else if (func->getKind() == FunctionConsts::OP_VAR_ASSIGN_1)
  {
    for (ulong i = 0; i < numArgs; ++i)
    {
      if (theArgs[i] == NULL)
        continue;

      expr_script_kind_t argKind = theArgs[i]->get_scripting_kind();

      if (argKind == UPDATE_EXPR)
      {
        ZORBA_ERROR_LOC(XUST0001, theArgs[i]->get_loc());
      }
    }

    theScriptingKind = SEQUENTIAL_EXPR;
  }
  else
  {
    theScriptingKind = func->getUpdateType();

    for (ulong i = 0; i < numArgs; ++i)
    {
      if (theArgs[i] == NULL)
        continue;

      expr_script_kind_t argKind = theArgs[i]->get_scripting_kind();

      if (argKind == UPDATE_EXPR)
      {
        ZORBA_ERROR_LOC(XUST0001, theArgs[i]->get_loc());
      }
      if (argKind == SEQUENTIAL_EXPR)
      {
        ZORBA_ERROR_LOC_DESC(XUST0001, theArgs[i]->get_loc(),
            "A sequential expression is not allowed at this position.");
      }
    }
  }
}


expr_t fo_expr::clone(substitution_t& subst) const
{
  if (get_func()->getKind() == FunctionConsts::FN_ZORBA_DDL_COLLECTION_1)
  {
    expr::subst_iter_t i = subst.find(this);

     if (i != subst.end())
       return i->second;
  }

  std::auto_ptr<fo_expr> fo(new fo_expr(theSctx, get_loc(), get_func()));

  for (unsigned i = 0; i < theArgs.size(); ++i)
    fo->theArgs.push_back(theArgs[i]->clone(subst));

  fo->theScriptingKind  = theScriptingKind;

  return fo.release();
}


}

