/*
 * Copyright 2006-2008 The FLWOR Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef ZORBA_SPECIFICATON
#define ZORBA_SPECIFICATON

#include <vector>
#include <string>
#include <fstream>
#include <sstream>


#undef WIN32_LEAN_AND_MEAN

class Specification
{
  typedef char char_t;
  typedef std::string::iterator iterator_t;

public:
  struct Variable 
  {
    bool        theInline;
    std::string theVarName;
    std::string theVarValue;
  };

  Specification()
    : theInline(false),
      theComparisonMethod("Fragment"),
      theUseIndent(false) 
  {}

private:
  bool                     theInline;
  std::string              theVarName;
  std::string              theVarValue;
  std::vector<Variable>    theVariables;
  std::vector<std::string> theResultFiles;
  std::vector<std::string> theErrors;
  std::string              theDate;
  std::string              theTimezone;
  std::string              theInputQueryFile;
  std::string              theComparisonMethod; // default is Fragment such that the user doesn't need to care about root tags for sequences etc
  std::string              theDefaultCollection;
  bool                     theUseIndent;

  void setInline() {
    theInline = true;
  }

  void setVarName(iterator_t str, iterator_t end) {
    theVarName = std::string(str, end);
  }

  void setVarValue(iterator_t str, iterator_t end) {
    theVarValue = std::string(str, end);
  }

  void addVariable() {
    Variable args = { theInline, theVarName, theVarValue};
    theVariables.push_back(args);
  }

  void addError(iterator_t str, iterator_t end) {
    std::string s(str, end);
    theErrors.push_back(s);
  }

  void setDate(iterator_t str, iterator_t end) {
    std::string s(str, end);
    theDate = s;
  }

  void setTimezone(iterator_t str, iterator_t end) {
    std::string s(str, end);
    theTimezone = s;
  }

  void setUseIndent() {
    theUseIndent = true;
  }

public:
  std::vector<Variable>::const_iterator
  variablesBegin() const { return theVariables.begin(); }

  std::vector<Variable>::const_iterator
  variablesEnd() const { return theVariables.end(); }

  std::vector<std::string>::const_iterator
  errorsBegin() const { return theErrors.begin(); }

  std::vector<std::string>::const_iterator
  errorsEnd() const { return theErrors.end(); }

  std::vector<std::string>::const_iterator
  resultsBegin() const { return theResultFiles.begin(); }

  std::vector<std::string>::const_iterator
  resultsEnd() const { return theResultFiles.end(); }

  size_t
  variablesSize() const { return theVariables.size(); }

  size_t
  errorsSize() const { return theErrors.size(); }

  bool hasDateSet() const {
    return theDate.size() != 0;
  }

  bool hasTimezoneSet() const {
    return theTimezone.size() != 0;
  }

  bool hasInputQuery () const {
    return theInputQueryFile.size() != 0;
  }

  const std::string & getInputQueryFile () const {
    return theInputQueryFile;
  }

  std::string getDate() const {
    return theDate;
  }

  std::string getTimezone() const {
    return theTimezone;
  }

  std::string getComparisonMethod() const {
    return theComparisonMethod;
  }

  std::string getDefaultCollection() const {
    return theDefaultCollection;
  }

  bool getUseIndent() const {
    return theUseIndent;
  }

  void tokenize(const std::string& str,
                std::vector<std::string>& tokens,
                const std::string& delimiters)
  {
    // Skip delimiters at beginning.
    std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    std::string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (std::string::npos != pos || std::string::npos != lastPos)
    {
      // Found a token, add it to the vector.
      tokens.push_back(str.substr(lastPos, pos - lastPos));
      // Skip delimiters.  Note the "not_of"
      lastPos = str.find_first_not_of(delimiters, pos);
      // Find next "non-delimiter"
      pos = str.find_first_of(delimiters, lastPos);
    }
  }

  void trim(std::string& str) {
    std::string::size_type  notwhite = str.find_first_not_of(" \t\n\r");
    str.erase(0,notwhite);

    notwhite = str.find_last_not_of(" \t\n\r"); 
    str.erase(notwhite+1); 
  }

  bool parseFile(std::string str)
  {
    std::ifstream lFile(str.c_str(), std::ifstream::in);
    std::stringstream lContent;
    while(lFile.good()){
      lContent << (char)lFile.get();
    }
    lFile.close();

    std::vector<std::string> lines;
    std::vector<std::string>::iterator it;

    tokenize(lContent.str(), lines, "\n");

    for(it = lines.begin(); it != lines.end(); ++it)
    {
      std::vector<std::string> tokens;
      std::vector<std::string>::iterator lIter;
      trim(*it);

      tokenize(*it, tokens, " ");

      for(lIter = tokens.begin(); lIter != tokens.end(); ++lIter)
      {
        if( *lIter == "Args:" )
        {
          for(++lIter; lIter!=tokens.end(); ++lIter)
          {
            if(*lIter != "-x"){ return false; }
            ++lIter;
            if(lIter->find(":=") == std::string::npos){ setInline(); }
            if(lIter->find('=') == std::string::npos){ return false; }
            if( theInline ){
              setVarName(lIter->begin(), lIter->begin()+lIter->find('='));
            } else {
              setVarName(lIter->begin(), lIter->begin()+lIter->find(":="));
            }
            setVarValue(lIter->begin()+lIter->find("=")+1, lIter->end());
            addVariable();
          }
          break;
        }
        else if ( *lIter == "Serialization:" )
        {
          for(++lIter; lIter!=tokens.end(); ++lIter)
          {
            trim(*lIter);
            // Just --indent for now
            if (*lIter == "--indent") {
              setUseIndent();
            } else {
              return false;
            }
          }
          break;
        }
        else if ( *lIter == "Result:" ) 
        {
          for (++lIter; lIter != tokens.end(); ++lIter)
          {
            trim(*lIter);
            theResultFiles.push_back(*lIter);
          }
          break;
        }
        else if ( *lIter == "InputQuery:" ) 
        {
          ++lIter;
          theInputQueryFile = *lIter;
        }
        else if ( *lIter == "Comparison:" ) 
        {
          ++lIter;
          theComparisonMethod = *lIter;
        }
        else if ( *lIter == "DefaultCollection:" ) 
        {
          ++lIter;
          theDefaultCollection = *lIter;
        }
        else if ( *lIter == "Error:" ) 
        { 
          ++lIter;
          if(lIter == tokens.end() ) { return false; }
          addError(lIter->begin(), lIter->end());
        }
        else if( *lIter == "Date:" ) 
        {
          ++lIter;
          if(lIter == tokens.end() ) { return false; }
          setDate( lIter->begin(), lIter->end() );
        }
        else if ( *lIter == "Timezone:" ) 
        {
          ++lIter;
          if(lIter == tokens.end() ) { return false; }
          setTimezone(lIter->begin(), lIter->end());
        }
        else if ( lIter->find("Error:") != std::string::npos ) 
        {
          addError(lIter->begin()+lIter->find(":")+1, lIter->end());
        }
        else if ( lIter->find("Date:") != std::string::npos  ) 
        {
          setDate(lIter->begin()+lIter->find(":")+1, lIter->end());
        }
        else if ( lIter->find("Timezone:") != std::string::npos ) 
        {
          setTimezone(lIter->begin()+lIter->find(":")+1, lIter->end());
        }
        else
        {
          break;
        }
      }
    }
    return true;
  }
};
#endif
